from time import time
from typing import List
import pandas as pd
import numpy as np
from sklearn.preprocessing import LabelEncoder, OneHotEncoder
from sklearn.base import ClusterMixin, BaseEstimator, BaseEstimator, ClassifierMixin
import pulp

class ClusterMilClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, clustering: ClusterMixin, weight_list: np.ndarray, onehot_encoder: OneHotEncoder):
        assert len(weight_list) > 0

        self.weight_list = weight_list
        self.clustering = clustering
        self.onehot_encoder = onehot_encoder
    def fit(self, X, y):
        # do nothing
        return self
    def predict(self, X):
        # bag_cluster_table (bag * cluster) @ weight (cluster*class)        
        new_cluster = self.clustering.predict(X)
        new_onehot_encoded_class = self.onehot_encoder.transform(np.array([new_cluster]).T)
        new_onehot_encoded_class.shape
        
        # its shape is (n_instance * n_class)
        n_instance = len(X)
        n_class = self.weight_list[0].shape[1]
        estimated_class_scores_sum = np.zeros((n_instance, n_class), dtype="float64")
        for i_W, W in enumerate(self.weight_list):
            estimated_class_scores = new_onehot_encoded_class @ W
            estimated_class_scores_sum += estimated_class_scores

        pred_y = np.argmax(estimated_class_scores_sum, axis=1)
        return pred_y


def generate_mil_classifier(
        clustering: ClusterMixin,
        cluster_encoder: OneHotEncoder,
        bags: List[np.ndarray],
        lower_threshold: np.ndarray,
        upper_threshold: np.ndarray,
        n_clusters: int,
        debug: bool = False):
    
    one_hot_encoded = [
        cluster_encoder.transform(
            clustering.predict(b).reshape((-1, 1))) for b in bags]
    
    bag_cluster_table = np.array(
        [np.sum(x, axis=0) for x in one_hot_encoded])
    
    n_classes = lower_threshold.shape[1]

    problem = pulp.LpProblem("clustermil", pulp.LpMinimize)
    margin = pulp.LpVariable("margin", lowBound=0)
    problem += margin

    W = []
    for i_class in range(n_classes):
        w = []
        for i_cluster in range(n_clusters):
            w.append(pulp.LpVariable(
                f"w_{i_class}_{i_cluster}", cat=pulp.LpBinary))
        W.append(w)
    for i_cluster in range(n_clusters):
        problem += pulp.lpSum(W[i_class][i_cluster] for i_class in range(n_classes)) == 1
    
    for i_bag in range(len(bags)):
        for i_class in range(n_classes):
            sum = None
            for i_cluster in range(n_clusters):
                if sum is None:
                    sum = W[i_class][i_cluster] * bag_cluster_table[i_bag][i_cluster]
                else:
                    sum += W[i_class][i_cluster] * bag_cluster_table[i_bag][i_cluster]
            
            problem += sum - margin <= upper_threshold[i_bag][i_class]
            problem += sum + margin >= lower_threshold[i_bag][i_class]
    
    
    problem.solve()
    W_float = np.array([np.array([x.value() for x in w]) for w in W]).T
    estimated_W_list = [W_float]

    if debug:
        print(f"W_float: {W_float}")
        print(f"margin: {margin.value()}")

    return ClusterMilClassifier(clustering, estimated_W_list, cluster_encoder)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['clustermil']

package_data = \
{'': ['*']}

install_requires = \
['PuLP>=2.6.0,<3.0.0',
 'numpy>=1.23.1,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'scikit-learn>=1.1.1,<2.0.0']

setup_kwargs = {
    'name': 'clustermil',
    'version': '0.2.0',
    'description': 'clustermil - clustering based multiple instance learning',
    'long_description': '# clustermil\n\n[![Build Status](https://app.travis-ci.com/inoueakimitsu/clustermil.svg?branch=main)](https://app.travis-ci.com/inoueakimitsu/clustermil)\n<a href="https://github.com/inoueakimitsu/clustermil/issues"><img alt="GitHub issues" src="https://img.shields.io/github/issues/inoueakimitsu/clustermil"></a> \n\nPython package for multiple instance learning (MIL) for large n_instance dataset.\n\n## Features\n\n- support count-based multiple instance assumptions (see [wikipedia](https://en.wikipedia.org/wiki/Multiple_instance_learning#:~:text=Presence-%2C%20threshold-%2C%20and%20count-based%20assumptions%5Bedit%5D))\n- support multi-class setting\n- support scikit-learn Clustering algorithms (such as `MiniBatchKMeans`)\n- fast even if n_instance is large\n\n## Installation\n\n```bash\npip install clustermil\n```\n\n## Usage\n\n```python\n# Prepare follwing dataset\n#\n# - bags ... list of np.ndarray\n#            (num_instance_in_the_bag * num_features)\n# - lower_threshold ... np.ndarray (num_bags * num_classes)\n# - upper_threshold ... np.ndarray (num_bags * num_classes)\n#\n# bags[i_bag] contains not less than lower_thrshold[i_bag, i_class]\n# i_class instances.\n\n# Prepare single-instance clustering algorithms\nfrom sklearn.cluster import MiniBatchKMeans\nn_clusters = 100\nclustering = MiniBatchKMeans(n_clusters=n_clusters)\nclusters = clustering.fit_predict(np.vstack(bags)) # flatten bags into instances\n\n# Prepare one-hot encoder\nfrom sklearn.preprocessing import OneHotEncoder\nonehot_encoder = OneHotEncoder()\nonehot_encoder.fit(clusters)\n\n# generate ClusterMilClassifier with helper function\nfrom clustermil import generate_mil_classifier\n\nmilclassifier = generate_mil_classifier(\n            clustering,\n            onehot_encoder,\n            bags,\n            lower_threshold,\n            upper_threshold,\n            n_clusters)\n\n# after multiple instance learning,\n# you can predict instance class\nmilclassifier.predict([instance_feature])\n```\n\nSee `tests/test_classification.py` for an example of a fully working test data generation process.\n\n## License\n\nclustermil is available under the MIT License.\n',
    'author': 'Inoue Akimitsu',
    'author_email': 'inoue.akimitsu@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/inoueakimitsu/clustermil',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
__version__.py
~~~~~~~~~~~~~~

Information about the current version of the clustermil package.
"""

__title__ = 'clustermil'
__description__ = 'clustermil - clustering based multiple instance learning'
__version__ = '0.2.0'
__author__ = 'Akimitsu Inoue'
__author_email__ = 'akimitsu.inoue@gmail.com'
__license__ = 'MIT'
__url__ = 'https://github.com/inoueakimitsu/clustermil'
